import asyncio

from blaxel.jobs import bl_job


async def main():
    job = bl_job("myjob")
    print(await job.arun([{"name": "charlou", "age": 25}]))
    print(job.run([{"name": "charlou", "age": 25}]))

if __name__ == "__main__":
    asyncio.run(main())



from typing import Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="JobsTasks")


@_attrs_define
class JobsTasks:
    """Jobs tasks

    Attributes:
        failed (Union[Unset, int]): Failed executions
        running (Union[Unset, int]): Running executions
        success (Union[Unset, int]): Success executions
        total (Union[Unset, int]): Total executions
    """

    failed: Union[Unset, int] = UNSET
    running: Union[Unset, int] = UNSET
    success: Union[Unset, int] = UNSET
    total: Union[Unset, int] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        failed = self.failed

        running = self.running

        success = self.success

        total = self.total

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if failed is not UNSET:
            field_dict["failed"] = failed
        if running is not UNSET:
            field_dict["running"] = running
        if success is not UNSET:
            field_dict["success"] = success
        if total is not UNSET:
            field_dict["total"] = total

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        if not src_dict:
            return None
        d = src_dict.copy()
        failed = d.pop("failed", UNSET)

        running = d.pop("running", UNSET)

        success = d.pop("success", UNSET)

        total = d.pop("total", UNSET)

        jobs_tasks = cls(
            failed=failed,
            running=running,
            success=success,
            total=total,
        )

        jobs_tasks.additional_properties = d
        return jobs_tasks

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="File")


@_attrs_define
class File:
    """
    Attributes:
        group (Union[Unset, str]):
        last_modified (Union[Unset, str]):
        name (Union[Unset, str]):
        owner (Union[Unset, str]):
        path (Union[Unset, str]):
        permissions (Union[Unset, str]):
        size (Union[Unset, int]):
    """

    group: Union[Unset, str] = UNSET
    last_modified: Union[Unset, str] = UNSET
    name: Union[Unset, str] = UNSET
    owner: Union[Unset, str] = UNSET
    path: Union[Unset, str] = UNSET
    permissions: Union[Unset, str] = UNSET
    size: Union[Unset, int] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        group = self.group

        last_modified = self.last_modified

        name = self.name

        owner = self.owner

        path = self.path

        permissions = self.permissions

        size = self.size

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if group is not UNSET:
            field_dict["group"] = group
        if last_modified is not UNSET:
            field_dict["lastModified"] = last_modified
        if name is not UNSET:
            field_dict["name"] = name
        if owner is not UNSET:
            field_dict["owner"] = owner
        if path is not UNSET:
            field_dict["path"] = path
        if permissions is not UNSET:
            field_dict["permissions"] = permissions
        if size is not UNSET:
            field_dict["size"] = size

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        if not src_dict:
            return None
        d = src_dict.copy()
        group = d.pop("group", UNSET)

        last_modified = d.pop("lastModified", UNSET)

        name = d.pop("name", UNSET)

        owner = d.pop("owner", UNSET)

        path = d.pop("path", UNSET)

        permissions = d.pop("permissions", UNSET)

        size = d.pop("size", UNSET)

        file = cls(
            group=group,
            last_modified=last_modified,
            name=name,
            owner=owner,
            path=path,
            permissions=permissions,
            size=size,
        )

        file.additional_properties = d
        return file

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""
This module provides a custom colored formatter for logging and an initialization function
to set up logging configurations for Blaxel applications.
"""
import json
import logging
import os

from opentelemetry.trace import get_current_span


class JsonFormatter(logging.Formatter):
    """
    A logger compatible with standard json logging.
    """
    def __init__(self):
        super().__init__()
        self.trace_id_name = os.environ.get('BL_LOGGER_TRACE_ID', 'trace_id')
        self.span_id_name = os.environ.get('BL_LOGGER_SPAN_ID', 'span_id')
        self.labels_name = os.environ.get('BL_LOGGER_LABELS', 'labels')
        self.trace_id_prefix = os.environ.get('BL_LOGGER_TRACE_ID_PREFIX', '')
        self.span_id_prefix = os.environ.get('BL_LOGGER_SPAN_ID_PREFIX', '')
        self.task_index = os.environ.get('BL_TASK_KEY', 'TASK_INDEX')
        self.task_prefix = os.environ.get('BL_TASK_PREFIX', '')
        self.execution_key = os.environ.get('BL_EXECUTION_KEY', 'BL_EXECUTION_ID')
        self.execution_prefix = os.environ.get('BL_EXECUTION_PREFIX', '')

    def format(self, record):
        """
        Formats the log record by converting it to a JSON object with trace context and environment variables.
        """
        log_entry = {
            'message': record.getMessage(),
            'severity': record.levelname,
            self.labels_name: {}
        }

        # Get current active span - equivalent to trace.getActiveSpan() in JS
        current_span = get_current_span()
        
        # Check if span exists and has valid context (equivalent to 'if (currentSpan)' in JS)
        if current_span and current_span.get_span_context().is_valid:
            span_context = current_span.get_span_context()
            # Format trace_id and span_id as hex strings (like JS does)
            trace_id_hex = format(span_context.trace_id, '032x')
            span_id_hex = format(span_context.span_id, '016x')
            
            log_entry[self.trace_id_name] = f"{self.trace_id_prefix}{trace_id_hex}"
            log_entry[self.span_id_name] = f"{self.span_id_prefix}{span_id_hex}"

        # Add task ID if available
        task_id = os.environ.get(self.task_index)
        if task_id:
            log_entry[self.labels_name]['blaxel-task'] = f"{self.task_prefix}{task_id}"

        # Add execution ID if available
        execution_id = os.environ.get(self.execution_key)
        if execution_id:
            log_entry[self.labels_name]['blaxel-execution'] = f"{self.execution_prefix}{execution_id.split('-')[-1]}"

        return json.dumps(log_entry)


class ColoredFormatter(logging.Formatter):
    """
    A custom logging formatter that adds ANSI color codes to log levels for enhanced readability.

    Attributes:
        COLORS (dict): A mapping of log level names to their corresponding ANSI color codes.
    """
    COLORS = {
        "DEBUG": "\033[1;36m",  # Cyan
        "INFO": "\033[1;32m",  # Green
        "WARNING": "\033[1;33m",  # Yellow
        "ERROR": "\033[1;31m",  # Red
        "CRITICAL": "\033[1;41m",  # Red background
    }

    def format(self, record):
        """
        Formats the log record by adding color codes based on the log level.

        Parameters:
            record (LogRecord): The log record to format.

        Returns:
            str: The formatted log message with appropriate color codes.
        """
        n_spaces = len("CRITICAL") - len(record.levelname)
        tab = " " * n_spaces
        color = self.COLORS.get(record.levelname, "\033[0m")
        record.levelname = f"{color}{record.levelname}\033[0m:{tab}"
        return super().format(record)

def init_logger(log_level: str):
    """
    Initializes the logging configuration for Blaxel.

    This function clears existing handlers for specific loggers, sets up a colored formatter,
    and configures the root logger with the specified log level.

    Parameters:
        log_level (str): The logging level to set (e.g., "DEBUG", "INFO").
    """
    # Disable urllib3 logging
    logging.getLogger('urllib3').setLevel(logging.CRITICAL)
    logging.getLogger("httpx").setLevel(logging.CRITICAL)
    handler = logging.StreamHandler()

    logger_type = os.environ.get("BL_LOGGER", "http")
    if logger_type == "json":
        handler.setFormatter(JsonFormatter())
    else:
        handler.setFormatter(ColoredFormatter("%(levelname)s %(name)s - %(message)s"))
    logging.basicConfig(level=log_level, handlers=[handler])



import os
from logging import getLogger

logger = getLogger(__name__)

os.environ["BL_FUNCTION_ADD_URL"] = "http://localhost:8080"

from blaxel.tools import bl_tools


async def main():
    async with bl_tools(["add"]) as t:
        tools = t.to_langchain()
        result = await tools[0].ainvoke({"a": 1, "b": 2})
        logger.info(result)

if __name__ == "__main__":
    import asyncio
    asyncio.run(main())



from blaxel.common.internal import get_alphanumeric_limited_hash

print(get_alphanumeric_limited_hash("charlou-function-blaxel-search", 48))

print(get_alphanumeric_limited_hash("main-function-explorer-mcp", 48))


from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.file import File
    from ..models.subdirectory import Subdirectory


T = TypeVar("T", bound="Directory")


@_attrs_define
class Directory:
    """
    Attributes:
        files (Union[Unset, list['File']]):
        name (Union[Unset, str]):
        path (Union[Unset, str]):
        subdirectories (Union[Unset, list['Subdirectory']]): @name Subdirectories
    """

    files: Union[Unset, list["File"]] = UNSET
    name: Union[Unset, str] = UNSET
    path: Union[Unset, str] = UNSET
    subdirectories: Union[Unset, list["Subdirectory"]] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        files: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.files, Unset):
            files = []
            for files_item_data in self.files:
                if type(files_item_data) == dict:
                    files_item = files_item_data
                else:
                    files_item = files_item_data.to_dict()
                files.append(files_item)

        name = self.name

        path = self.path

        subdirectories: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.subdirectories, Unset):
            subdirectories = []
            for subdirectories_item_data in self.subdirectories:
                if type(subdirectories_item_data) == dict:
                    subdirectories_item = subdirectories_item_data
                else:
                    subdirectories_item = subdirectories_item_data.to_dict()
                subdirectories.append(subdirectories_item)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if files is not UNSET:
            field_dict["files"] = files
        if name is not UNSET:
            field_dict["name"] = name
        if path is not UNSET:
            field_dict["path"] = path
        if subdirectories is not UNSET:
            field_dict["subdirectories"] = subdirectories

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.file import File
        from ..models.subdirectory import Subdirectory

        if not src_dict:
            return None
        d = src_dict.copy()
        files = []
        _files = d.pop("files", UNSET)
        for files_item_data in _files or []:
            files_item = File.from_dict(files_item_data)

            files.append(files_item)

        name = d.pop("name", UNSET)

        path = d.pop("path", UNSET)

        subdirectories = []
        _subdirectories = d.pop("subdirectories", UNSET)
        for subdirectories_item_data in _subdirectories or []:
            subdirectories_item = Subdirectory.from_dict(subdirectories_item_data)

            subdirectories.append(subdirectories_item)

        directory = cls(
            files=files,
            name=name,
            path=path,
            subdirectories=subdirectories,
        )

        directory.additional_properties = d
        return directory

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class ProcessResponseStatus(str, Enum):
    COMPLETED = "completed"
    FAILED = "failed"
    KILLED = "killed"
    RUNNING = "running"
    STOPPED = "stopped"

    def __str__(self) -> str:
        return str(self.value)



"""Contains types for authentication credentials"""

from typing import Dict, Optional

from httpx import Auth
from pydantic import BaseModel, Field


class CredentialsType(BaseModel):
    """Represents authentication credentials for the API"""

    api_key: Optional[str] = Field(default=None, description="The API key")
    client_credentials: Optional[str] = Field(default=None, description="The client credentials")
    refresh_token: Optional[str] = Field(default=None, description="The refresh token")
    access_token: Optional[str] = Field(default=None, description="The access token")
    device_code: Optional[str] = Field(default=None, description="The device code")
    expires_in: Optional[int] = Field(default=None, description="The expiration time")
    workspace: Optional[str] = Field(default=None, description="The workspace")

class BlaxelAuth(Auth):
    def __init__(self, credentials: CredentialsType, workspace_name: str, base_url: str):
        """
        Initializes the BlaxelAuth with the given credentials, workspace name, and base URL.

        Parameters:
            credentials: Credentials containing the Bearer token and refresh token.
            workspace_name (str): The name of the workspace.
            base_url (str): The base URL for authentication.
        """
        self.credentials = credentials
        self.workspace_name = workspace_name
        self.base_url = base_url

    def get_headers(self) -> Dict[str, str]:
        return {}

    @property
    def token(self):
        raise NotImplementedError("Subclasses must implement the token property")



import asyncio

from dotenv import load_dotenv

load_dotenv()

from logging import getLogger

from pydantic_ai.messages import ModelRequest, UserPromptPart
from pydantic_ai.models import ModelRequestParameters, ModelSettings

from blaxel.models import bl_model

logger = getLogger(__name__)


MODEL = "gpt-4o-mini"
# MODEL = "claude-3-5-sonnet"
# MODEL = "xai-grok-beta"
# MODEL = "cohere-command-r-plus"
# MODEL = "gemini-2-0-flash"
# MODEL = "deepseek-chat"
# MODEL = "mistral-large-latest"
# MODEL = "cerebras-llama-4-scout-17b"

async def test_model_langchain():
    model = await bl_model(MODEL).to_langchain()
    result = await model.ainvoke("Hello, world!")
    logger.info(result)

async def test_model_llamaindex():
    model = await bl_model(MODEL).to_llamaindex()
    result = await model.acomplete("Hello, world!")
    logger.info(result)

async def test_model_crewai():
    # not working: cohere
    model = await bl_model(MODEL).to_crewai()
    result = model.call([{"role": "user", "content": "Hello, world!"}])
    logger.info(result)

async def test_model_pydantic():
    model = await bl_model(MODEL).to_pydantic()
    result = await model.request(
        [ModelRequest(parts=[UserPromptPart(content="Hello, world!")])],
        model_settings=ModelSettings(max_tokens=100),
        model_request_parameters=ModelRequestParameters(function_tools=[], allow_text_result=True, result_tools=[])
    )
    logger.info(result)

async def test_model_google_adk():
    from google.adk.models.llm_request import LlmRequest

    model = await bl_model(MODEL).to_google_adk()
    request = LlmRequest(
        model=MODEL,
        contents=[{"role": "user", "parts": [{"text": "Hello, world!"}]}],
        config={},
        tools_dict={},
    )
    async for result in model.generate_content_async(request):
        logger.info(result)

async def main():
    # await test_model_langchain()
    # await test_model_llamaindex()
    # await test_model_crewai()
    # await test_model_pydantic()
    await test_model_google_adk()

if __name__ == "__main__":
    asyncio.run(main())


import logging
import traceback
import uuid
from contextlib import asynccontextmanager
from typing import Dict, Literal

import anyio
import mcp.types as types
from anyio.streams.memory import MemoryObjectReceiveStream, MemoryObjectSendStream
from mcp.server.fastmcp import FastMCP as FastMCPBase
from opentelemetry.trace import Span, StatusCode
from websockets.server import WebSocketServerProtocol, serve

from ..common.env import env
from ..instrumentation.span import SpanManager

logger = logging.getLogger(__name__)


class BlaxelMcpServerTransport:
    """WebSocket server transport for MCP."""
    spans: Dict[str, Span] = {}

    def __init__(self, port: int = 8080):
        """Initialize the WebSocket server transport.

        Args:
            port: The port to listen on (defaults to 8080 or BL_SERVER_PORT env var)
        """
        if env["BL_SERVER_PORT"] is not None:
            self.port = int(env["BL_SERVER_PORT"])
        else:
            self.port = port
        self.clients = {}
        self.server = None
        self.span_manager = SpanManager("blaxel-tracer")

    @asynccontextmanager
    async def websocket_server(self):
        """Create and run a WebSocket server for MCP communication."""
        read_stream: MemoryObjectReceiveStream[types.JSONRPCMessage | Exception]
        read_stream_writer: MemoryObjectSendStream[types.JSONRPCMessage | Exception]

        write_stream: MemoryObjectSendStream[types.JSONRPCMessage]
        write_stream_reader: MemoryObjectReceiveStream[types.JSONRPCMessage]

        read_stream_writer, read_stream = anyio.create_memory_object_stream(0)
        write_stream, write_stream_reader = anyio.create_memory_object_stream(0)

        async def handler(websocket: WebSocketServerProtocol):
            client_id = str(uuid.uuid4())
            self.clients[client_id] = websocket
            logger.info(f"Client connected: {client_id}")

            try:
                async for message in websocket:
                    span = self.span_manager.create_span("message", {"mcp.client.id": client_id})
                    try:
                        msg = types.JSONRPCMessage.model_validate_json(message)
                        # Modify message ID to include client ID
                        if hasattr(msg, "id") and msg.id is not None:
                            original_id = msg.id
                            msg.id = f"{client_id}:{original_id}"
                            span.set_attributes({
                                "mcp.message.parsed": True,
                                "mcp.method": getattr(msg, "method", None),
                                "mcp.messageId": getattr(msg, "id", None),
                                "mcp.toolName": getattr(getattr(msg, "params", None), "name", None),
                                "span.type": "mcp.message",
                            })
                            self.spans[client_id+":"+msg.id] =  span
                        await read_stream_writer.send(msg)
                    except Exception as exc:
                        span.set_status(StatusCode.ERROR)
                        span.record_exception(exc)
                        span.end()
                        logger.error(f"Failed to parse message: {exc}\n{traceback.format_exc()}")
                        await read_stream_writer.send(exc)
            except Exception as e:
                logger.error(f"WebSocket error: {e}")
            finally:
                if client_id in self.clients:
                    del self.clients[client_id]
                logger.info(f"Client disconnected: {client_id}")

        async def message_sender():
            async with write_stream_reader:
                async for message in write_stream_reader:
                    # Extract client ID from message ID
                    client_id = None
                    msg_id = None

                    if hasattr(message, "id") and message.id is not None:
                        parts = str(message.id).split(":", 1)
                        if len(parts) == 2:
                            client_id, msg_id = parts
                            # Restore original message ID
                            message.id = int(msg_id) if msg_id.isdigit() else msg_id

                    data = message.model_dump_json(by_alias=True, exclude_none=True)

                    if client_id and client_id in self.clients:
                        # Send to specific client
                        websocket = self.clients[client_id]
                        span = self.spans.get(client_id+":"+msg_id)
                        try:
                            await websocket.send(data)
                            if span:
                                span.set_attributes({
                                    "mcp.message.response_sent": True,
                                })
                        except Exception as e:
                            if span:
                                span.set_status(StatusCode.ERROR)
                                span.record_exception(e)
                            logger.error(f"Failed to send message to client {client_id}: {e}")
                            if client_id in self.clients:
                                del self.clients[client_id]
                        finally:
                            if span:
                                span.end()
                    else:
                        # Broadcast to all clients
                        dead_clients = []
                        for cid, websocket in self.clients.items():
                            try:
                                await websocket.send(data)
                            except Exception:
                                dead_clients.append(cid)

                        # Clean up dead clients
                        for cid in dead_clients:
                            if cid in self.clients:
                                del self.clients[cid]

        async with anyio.create_task_group() as tg:
            logger.info(f"Starting WebSocket Server on port {self.port}")
            async with serve(handler, "0.0.0.0", self.port) as server:
                self.server = server
                tg.start_soon(message_sender)
                yield read_stream, write_stream

class FastMCP(FastMCPBase):
    def run(self, transport: Literal["stdio", "sse", "ws"] = "stdio") -> None:
        """Run the FastMCP server. Note this is a synchronous function.

        Args:
            transport: Transport protocol to use ("stdio" or "sse")
        """
        TRANSPORTS = Literal["stdio", "sse", "ws"]
        if transport not in TRANSPORTS.__args__:  # type: ignore
            raise ValueError(f"Unknown transport: {transport}")

        if transport == "stdio":
            anyio.run(self.run_stdio_async)
        elif transport == "ws":
            anyio.run(self.run_ws_async)
        else:
            anyio.run(self.run_sse_async)

    async def run_ws_async(self) -> None:
        async with BlaxelMcpServerTransport().websocket_server() as (read_stream, write_stream):
            await self._mcp_server.run(
                read_stream,
                write_stream,
                self._mcp_server.create_initialization_options(),
            )



from ..client import client
from ..instrumentation.manager import telemetry_manager
from .settings import settings


def autoload() -> None:
    client.with_base_url(settings.base_url)
    client.with_auth(settings.auth)
    telemetry_manager.initialize(settings)


import asyncio
import logging
import time
from typing import List

from ..client.api.compute.create_sandbox import asyncio as create_sandbox
from ..client.api.compute.delete_sandbox import asyncio as delete_sandbox
from ..client.api.compute.get_sandbox import asyncio as get_sandbox
from ..client.api.compute.list_sandboxes import asyncio as list_sandboxes
from ..client.client import client
from ..client.models import Sandbox
from .filesystem import SandboxFileSystem
from .preview import SandboxPreviews
from .process import SandboxProcess

logger = logging.getLogger(__name__)

class SandboxInstance:
    def __init__(self, sandbox: Sandbox):
        self.sandbox = sandbox
        self.fs = SandboxFileSystem(sandbox)
        self.process = SandboxProcess(sandbox)
        self.previews = SandboxPreviews(sandbox)

    @property
    def metadata(self):
        return self.sandbox.metadata

    @property
    def status(self):
        return self.sandbox.status

    @property
    def events(self):
        return self.sandbox.events

    @property
    def spec(self):
        return self.sandbox.spec

    async def wait(self, max_wait: int = 60000, interval: int = 1000) -> None:
        start_time = time.time() * 1000  # Convert to milliseconds
        while self.sandbox.status != "DEPLOYED":
            await asyncio.sleep(interval / 1000)  # Convert to seconds
            try:
                response = await get_sandbox(
                    self.sandbox.metadata.name,
                    client=client,
                )
                logger.info(f"Waiting for sandbox to be deployed, status: {response.status}")
                self.sandbox = response
            except Exception as e:
                logger.error("Could not retrieve sandbox", exc_info=e)

            if self.sandbox.status == "FAILED":
                raise Exception("Sandbox failed to deploy")

            if (time.time() * 1000) - start_time > max_wait:
                raise Exception("Sandbox did not deploy in time")

    @classmethod
    async def create(cls, sandbox: Sandbox) -> "SandboxInstance":
        if not sandbox.spec:
            raise Exception("Sandbox spec is required")
        if not sandbox.spec.runtime:
            raise Exception("Sandbox runtime is required")
        sandbox.spec.runtime.generation = sandbox.spec.runtime.generation or "mk3"

        response = await create_sandbox(
            client=client,
            body=sandbox,
        )
        return cls(response)

    @classmethod
    async def get(cls, sandbox_name: str) -> "SandboxInstance":
        response = await get_sandbox(
            sandbox_name,
            client=client,
        )
        return cls(response)

    @classmethod
    async def list(cls) -> List["SandboxInstance"]:
        response = await list_sandboxes()
        return [cls(sandbox) for sandbox in response]

    @classmethod
    async def delete(cls, sandbox_name: str) -> Sandbox:
        response = await delete_sandbox(
            sandbox_name,
            client=client,
        )
        return response



from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.request_total_by_origin_metric_request_total_by_origin import (
        RequestTotalByOriginMetricRequestTotalByOrigin,
    )
    from ..models.request_total_by_origin_metric_request_total_by_origin_and_code import (
        RequestTotalByOriginMetricRequestTotalByOriginAndCode,
    )


T = TypeVar("T", bound="RequestTotalByOriginMetric")


@_attrs_define
class RequestTotalByOriginMetric:
    """Request total by origin metric

    Attributes:
        request_total_by_origin (Union[Unset, RequestTotalByOriginMetricRequestTotalByOrigin]): Request total by origin
        request_total_by_origin_and_code (Union[Unset, RequestTotalByOriginMetricRequestTotalByOriginAndCode]): Request
            total by origin and code
    """

    request_total_by_origin: Union[Unset, "RequestTotalByOriginMetricRequestTotalByOrigin"] = UNSET
    request_total_by_origin_and_code: Union[Unset, "RequestTotalByOriginMetricRequestTotalByOriginAndCode"] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        request_total_by_origin: Union[Unset, dict[str, Any]] = UNSET
        if (
            self.request_total_by_origin
            and not isinstance(self.request_total_by_origin, Unset)
            and not isinstance(self.request_total_by_origin, dict)
        ):
            request_total_by_origin = self.request_total_by_origin.to_dict()
        elif self.request_total_by_origin and isinstance(self.request_total_by_origin, dict):
            request_total_by_origin = self.request_total_by_origin

        request_total_by_origin_and_code: Union[Unset, dict[str, Any]] = UNSET
        if (
            self.request_total_by_origin_and_code
            and not isinstance(self.request_total_by_origin_and_code, Unset)
            and not isinstance(self.request_total_by_origin_and_code, dict)
        ):
            request_total_by_origin_and_code = self.request_total_by_origin_and_code.to_dict()
        elif self.request_total_by_origin_and_code and isinstance(self.request_total_by_origin_and_code, dict):
            request_total_by_origin_and_code = self.request_total_by_origin_and_code

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if request_total_by_origin is not UNSET:
            field_dict["requestTotalByOrigin"] = request_total_by_origin
        if request_total_by_origin_and_code is not UNSET:
            field_dict["requestTotalByOriginAndCode"] = request_total_by_origin_and_code

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.request_total_by_origin_metric_request_total_by_origin import (
            RequestTotalByOriginMetricRequestTotalByOrigin,
        )
        from ..models.request_total_by_origin_metric_request_total_by_origin_and_code import (
            RequestTotalByOriginMetricRequestTotalByOriginAndCode,
        )

        if not src_dict:
            return None
        d = src_dict.copy()
        _request_total_by_origin = d.pop("requestTotalByOrigin", UNSET)
        request_total_by_origin: Union[Unset, RequestTotalByOriginMetricRequestTotalByOrigin]
        if isinstance(_request_total_by_origin, Unset):
            request_total_by_origin = UNSET
        else:
            request_total_by_origin = RequestTotalByOriginMetricRequestTotalByOrigin.from_dict(_request_total_by_origin)

        _request_total_by_origin_and_code = d.pop("requestTotalByOriginAndCode", UNSET)
        request_total_by_origin_and_code: Union[Unset, RequestTotalByOriginMetricRequestTotalByOriginAndCode]
        if isinstance(_request_total_by_origin_and_code, Unset):
            request_total_by_origin_and_code = UNSET
        else:
            request_total_by_origin_and_code = RequestTotalByOriginMetricRequestTotalByOriginAndCode.from_dict(
                _request_total_by_origin_and_code
            )

        request_total_by_origin_metric = cls(
            request_total_by_origin=request_total_by_origin,
            request_total_by_origin_and_code=request_total_by_origin_and_code,
        )

        request_total_by_origin_metric.additional_properties = d
        return request_total_by_origin_metric

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Awaitable, Callable, Dict, Optional

from pydantic import BaseModel


class ToolException(Exception):  # noqa: N818
    """Optional exception that tool throws when execution error occurs.

    When this exception is thrown, the agent will not stop working,
    but it will handle the exception according to the handle_tool_error
    variable of the tool, and the processing result will be returned
    to the agent as observation, and printed in red on the console.
    """

class Tool(BaseModel):
    name: str
    description: str
    input_schema: Dict[str, Any]
    coroutine: Optional[Callable[..., Awaitable[Any]]] = None
    sync_coroutine: Optional[Callable[..., Any]] = None